/*
 * Copyright (C) 2006 Quinn Storm <livinglatexkali@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <string.h>
#include <malloc.h>
#define USE_GETTEXT
#include <beryl.h>
#include <beryl-settings.h>
#include <stdlib.h>
#include<libintl.h>
#include<locale.h>
#include<signal.h>
#include<unistd.h>
#define _(String) gettext (String)
#define gettext_noop(String) String
#ifndef N_
#define N_(String) gettext_noop (String)
#endif

BerylSettingsContext * context;
gchar * fnprofile;
gchar * profilefile;
gchar * profilefilecontent;
gchar * profiles [] = {
    ".beryl/settings1.Profile",
    ".beryl/settings2.Profile",
    ".beryl/settings3.Profile",
    ".beryl/settings4.Profile",
    ".beryl/settings5.Profile"
};
GtkWidget * EntryBox;

typedef struct _SETNumericWidgets
{
    GtkSpinButton * s;
    GtkHScale * h;
} SETNumericWidgets;
gint nModifiers[]={
    ShiftMask,
    ControlMask,
    CompAltMask,
    CompSuperMask,
    CompHyperMask,
    CompMetaMask,
    CompModeSwitchMask,
    Mod1Mask,
    Mod2Mask,
    Mod3Mask,
    Mod4Mask,
    Mod5Mask,
};
gchar * modifiers[]={
    N_("Shift"),
    N_("Control"),
    N_("Alt"),
    N_("Super"),
    N_("Hyper"),
    N_("Meta"),
    N_("Mode"),
    N_("Mod1"),
    N_("Mod2"),
    N_("Mod3"),
    N_("Mod4"),
    N_("Mod5"),
};
#define NMODIFIERS (sizeof(modifiers)/sizeof(gchar *))
#define MODROWS 3
gchar * edges[]={
    N_("Left"),
    N_("Right"),
    N_("Top"),
    N_("Bottom"),
    N_("Top-Left"),
    N_("Top-Right"),
    N_("Bottom-Left"),
    N_("Bottom-Right"),
};
// intuitive layout of edge checkboxes in the table
gint edgecols[]= {
    0,
    2,
    1,
    1,
    0,
    2,
    0,
    2,
};
gint edgerows[]= {
    1,
    1,
    0,
    2,
    0,
    0,
    2,
    2,
};
typedef struct _SETActionWidgets
{
    GtkCheckButton  * kmod[NMODIFIERS];      // modifiers for key
    GtkCheckButton  * keyen;
    GtkEntry        * key;                   // keysym(string in entrybox)
    GtkCheckButton  * bmod[NMODIFIERS];      // modifiers for mouse button
    GtkCheckButton  * buten;
    GtkEntry        * but;                   // mouse button
    GtkCheckButton  * b;                     // bell
    GtkCheckButton  * e[SCREEN_EDGE_NUM];    // edges (checklist)
} SETActionWidgets;
typedef struct _SETStringListWidgets
{
    GtkListStore * backing;
    GtkTreeView  * view;
    GtkButton * add;
    GtkButton * rem;
    GtkButton * bro;
    GtkButton * up;
    GtkButton * dn;
} SETStringListWidgets;
typedef struct _SETStringBrowseWidgets
{
    GtkEntry * e;
    GtkButton * b;
} SETStringBrowseWidgets;
typedef union _SETOptionWidgets
{
    GtkWidget * w;
    GtkCheckButton * b;
    SETStringBrowseWidgets sb;
    SETNumericWidgets n;
    SETActionWidgets a;
    GtkWidget ** l;
    SETStringListWidgets s;
} SETOptionWidgets;
typedef struct _SETGroupWidget
{
    gchar * group;
    GtkWidget * widget;
    gboolean hasany;
} SETGroupWidget;
typedef struct _SETPage
{
    gchar * name;
    GtkWidget * widget;
} SETPage;
typedef struct _SETResetOption
{
    BerylSetting * o;
    gint resetType; // for action options, 0=button,1=key,2=edge,3=bell
} SETResetOption;
typedef struct _SETPlugin
{
    BerylSettingsPlugin * plugin;
    gboolean active;
} SETPlugin;
SETPage * SETOptionPages;
GSList * Plugins=NULL;
GSList * SETOptions=NULL;
GtkListStore * pluginList;
GtkTreeView * SETPluginList;
GtkWidget * pluginContainer;
GtkWidget * infoLabel;
GtkWidget * infoContainer;
gint nPages=0;
GtkWidget * mainWindow;
static gboolean suspend_signals=FALSE;
gint latestProfile=0;

enum _SETPluginListEntry
{
    PluginName,
    ShortDesc,
    LongDesc,
    Markup,
    PluginIcon,
    PluginEnabled,
    PluginCanEnable,
    PluginEntryCount
};

gboolean check_contains(GSList * hay, gchar * needle)
{
    GSList * stalk;
    for(stalk=hay;stalk;stalk=stalk->next)
    {
        if (!g_strcasecmp(stalk->data,needle))
            return TRUE;
    }
    return FALSE;
}
gboolean check_intersects(GSList * hay, GSList * pitchfork)
{
    GSList * tine;
    for(tine=pitchfork;tine;tine=tine->next)
    {
        if (check_contains(hay,tine->data))
            return TRUE;
    }
    return FALSE;
}

void queue_option_changed()
{
    beryl_settings_context_write(context);
    beryl_settings_send_reload_signal();
}
void cb_option(GtkWidget * w, BerylSetting * o)
{
    if (suspend_signals)
        return;
    SETOptionWidgets * widgets=o->private_ptr;
    gboolean b_val;
    gint i_val;
    switch(o->type)
    {
        case BERYL_SETTING_TYPE_BOOL:
            b_val=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets->b));
            beryl_setting_value_set_bool(&o->value,&b_val);
            break;
        case BERYL_SETTING_TYPE_INT:
        case BERYL_SETTING_TYPE_FLOAT:
            {
                gdouble f;
                if (GTK_WIDGET(w)==GTK_WIDGET(widgets->n.s))
                {
                    f=gtk_spin_button_get_value(GTK_SPIN_BUTTON(w));
                }
                else
                {
                    f=gtk_range_get_value(GTK_RANGE(w));
                }
                if (o->type==BERYL_SETTING_TYPE_INT)
                {
                    i_val=(gint)f;
                    beryl_setting_value_set_int(&o->value,&i_val);
                }
                else
                {
                    beryl_setting_value_set_float(&o->value,&f);
                }
                gtk_range_set_value(GTK_RANGE(widgets->n.h),f);
                gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets->n.s),f);
            }
            break;
        case BERYL_SETTING_TYPE_STRING:
            {
                const gchar * s;
                if (o->info.for_string.allowed_values)
                    s=gtk_combo_box_get_active_text(GTK_COMBO_BOX(widgets->w));
                else
                    s = gtk_entry_get_text(GTK_ENTRY(widgets->w));
                beryl_setting_value_set_string(&o->value,&s);
            }
            break;
        case BERYL_SETTING_TYPE_COLOR:
            {
                GdkColor c;
                gint a;
                BerylSettingColorValue cval;
                gtk_color_button_get_color(GTK_COLOR_BUTTON(widgets->w),&c);
                a=gtk_color_button_get_alpha(GTK_COLOR_BUTTON(widgets->w));
                cval.color.red=c.red;
                cval.color.green=c.green;
                cval.color.blue=c.blue;
                cval.color.alpha=a;
                beryl_setting_value_set_color(&o->value,&cval);
            }
            break;
        case BERYL_SETTING_TYPE_BINDING:
            {
                if (o->info.for_bind.button)
                {
                    gint mods=0;
                    gint i;
                    gboolean b;
                    b=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets->a.buten));
                    beryl_setting_value_set_button_enabled(&o->value,&b);
                    for(i=0;i<NMODIFIERS;i++)
                        if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets->a.bmod[i])))
                            mods|=nModifiers[i];
                    beryl_setting_value_set_buttonmods(&o->value,&mods);
                    mods=atoi(gtk_entry_get_text(GTK_ENTRY(widgets->a.but)));
                    beryl_setting_value_set_button(&o->value,&mods);
                }
                if (o->info.for_bind.key)
                {
                    gint mods=0;
                    gint i;
                    gboolean b;
                    b=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets->a.keyen));
                    beryl_setting_value_set_key_enabled(&o->value,&b);
                    for(i=0;i<NMODIFIERS;i++)
                        if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets->a.kmod[i])))
                            mods|=nModifiers[i];
                    beryl_setting_value_set_keymods(&o->value,&mods);
                    mods=XStringToKeysym(gtk_entry_get_text(GTK_ENTRY(widgets->a.key)));
                    beryl_setting_value_set_keysym(&o->value,&mods);
                }
                if (o->info.for_bind.edge)
                {
                    gint edg=0;
                    gint i;
                    for(i=0;i<SCREEN_EDGE_NUM;i++)
                    {
                        if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets->a.e[i])))
                            edg|=1<<i;
                    }
                    beryl_setting_value_set_edgemask(&o->value,&edg);
                }
                if (o->info.for_bind.bell)
                {
                    gboolean bel=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets->a.b));
                    beryl_setting_value_set_bell(&o->value,&bel);
                }
            }
            break;
        case BERYL_SETTING_TYPE_LIST:
            {
                if (o->info.for_list.list_of_type==BERYL_SETTING_TYPE_STRING && 
                        o->info.for_list.list_of_info->for_string.allowed_values)
                {
                    beryl_setting_list_clear(o);
                    //multiple-choice
                    gint i;
                    GSList * j=o->info.for_list.list_of_info->for_string.allowed_values;
                    for (i=0;j;j=j->next)
                    {
                        if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets->l[i])))
                        {
                            BerylSettingValue * val = beryl_setting_list_append(o);
                            const gchar * s = j->data;
                            beryl_setting_value_set_string(val,&s);
                        }
                        i++;
                    }
                }
            }
            break;
        default:
            break;
    }
    beryl_settings_context_write(context);
    beryl_settings_send_reload_signal();
}
void set_info_pane(GtkWidget * w, gchar * labelMarkup)
{
    if (strcmp(labelMarkup,"")==0)
    {
        gtk_widget_hide(infoContainer);
    }
    else
    {
        gtk_window_set_resizable(GTK_WINDOW(mainWindow),FALSE);
        gtk_label_set_markup(GTK_LABEL(infoLabel),labelMarkup);
        gtk_widget_show_all(infoContainer);
        gtk_window_set_resizable(GTK_WINDOW(mainWindow),TRUE);
    }
}
gchar * getmark(BerylSetting * o)
{
    return g_markup_printf_escaped(
            "<b>%s</b>\n"
            "<small><i>(%s/%s)</i></small>\n"
            "<small>%s</small>",
            beryl_setting_get_short_desc(o),
            o->parent->name?o->parent->name:"_",
            o->name,
            beryl_setting_get_long_desc(o)
            );
}
GtkWidget * my_label_new(const gchar * t)
{
    GtkWidget * w = gtk_label_new(t);
    gtk_misc_set_alignment(GTK_MISC(w),0,0);
    return w;
}
void set_plugin_active(gchar * plugin,gboolean active)
{
    GtkTreeIter it;
    gboolean k;
    GSList * pl;
    for(k=gtk_tree_model_get_iter_first(GTK_TREE_MODEL(pluginList),&it);k;
            k=gtk_tree_model_iter_next(GTK_TREE_MODEL(pluginList),&it))
    {
        gchar * pl;
        gtk_tree_model_get(GTK_TREE_MODEL(pluginList),&it,PluginName,&pl,-1);
        if (strcmp(pl,plugin)==0)
            gtk_list_store_set(pluginList,&it,PluginEnabled,active,-1);
        g_free(pl);
    }
    for(pl=Plugins;pl;pl=pl->next)
    {
        if (!g_strcasecmp(((SETPlugin *)pl->data)->plugin->name?
                    ((SETPlugin *)pl->data)->plugin->name:"_",plugin))
        {
            ((SETPlugin *)pl->data)->active=active;
            break;
        }
    }
}


gboolean doing_active_plugins(BerylSetting * o, gboolean reset)
{
    if (strcmp(o->name,"active_plugins")==0 && !o->parent->name)
    {
        //initialize to inactive
        if (reset)
        {
            GSList * pl;
            for (pl=Plugins;pl;pl=pl->next)
            {
                SETPlugin * p=pl->data;
                set_plugin_active(p->plugin->name,FALSE);
            }
        }
        //o is of type list[string]
        GSList * p;
        beryl_setting_value_get_value_list(&o->value,&p);
        for(;p;p=p->next)
        {
            //set the plugins active
            //by first finding them...
            set_plugin_active(((BerylSettingValue *)p->data)->value.as_string,TRUE);
        }
        return TRUE;
    }
    return FALSE;
}
void set_widgets_for_setting(BerylSetting * setting)
{
    suspend_signals=TRUE;
    SETOptionWidgets * widgets;
    gboolean b_val;
    GSList * l_val;
    gint i_val;
    gint i;
    gdouble f_val;
    const gchar * s_val;
    BerylSettingValue * value = beryl_setting_get_primary_value(setting);
    widgets=beryl_setting_get_private(setting);
    switch(beryl_setting_get_type(setting))
    {
        case BERYL_SETTING_TYPE_BOOL:
            if (beryl_setting_value_get_bool(value,&b_val))
                gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->w),b_val);
            break;
        case BERYL_SETTING_TYPE_STRING:
            if (beryl_setting_get_allowed_strings(setting,&l_val) && l_val)
            {
                i=0;
                for (;l_val;l_val=l_val->next)
                {
                    beryl_setting_value_get_string(value,&s_val);
                    if (strcmp(l_val->data,s_val)==0)
                        gtk_combo_box_set_active(GTK_COMBO_BOX(widgets->w),i);
                    i++;
                }
            }
            else
            {
                beryl_setting_value_get_string(value,&s_val);
                gtk_entry_set_text(GTK_ENTRY(widgets->w),s_val);
            }
            break;
        case BERYL_SETTING_TYPE_INT:
            beryl_setting_value_get_int(value,&i_val);
            gtk_range_set_value(GTK_RANGE(widgets->n.h),i_val);
            gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets->n.s),i_val);
            break;
        case BERYL_SETTING_TYPE_FLOAT:
            beryl_setting_value_get_float(value,&f_val);
            gtk_range_set_value(GTK_RANGE(widgets->n.h),f_val);
            gtk_spin_button_set_value(GTK_SPIN_BUTTON(widgets->n.s),f_val);
            break;
        case BERYL_SETTING_TYPE_COLOR:
            {
                BerylSettingColorValue cval;
                beryl_setting_value_get_color(value,&cval);
                GdkColor c;
                c.red=cval.array.array[0];
                c.green=cval.array.array[1];
                c.blue=cval.array.array[2];
                gtk_color_button_set_color(GTK_COLOR_BUTTON(widgets->w),&c);
                gtk_color_button_set_alpha(GTK_COLOR_BUTTON(widgets->w),
                        cval.array.array[3]);
            }
            break;
        case BERYL_SETTING_TYPE_BINDING:
            if (beryl_setting_get_can_set_button(setting,&b_val) && b_val)
            {
                beryl_setting_value_get_button_enabled(value,&b_val);
                gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->a.buten),b_val);
                gint i;
                beryl_setting_value_get_buttonmods(value,&i_val);
                for(i=0;i<NMODIFIERS;i++)
                {
                    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->a.bmod[i]),
                            (i_val & nModifiers[i]));
                }
                beryl_setting_value_get_button(value,&i_val);
                gchar * s = g_strdup_printf("%d",i_val);
                gtk_entry_set_text(GTK_ENTRY(widgets->a.but),s);
                g_free(s);
            }
            if (beryl_setting_get_can_set_key(setting,&b_val) && b_val)
            {
                gint i;
                beryl_setting_value_get_key_enabled(value,&b_val);
                gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->a.keyen),b_val);
                gchar * s;
                beryl_setting_value_get_keymods(value,&i_val);
                for(i=0;i<NMODIFIERS;i++)
                {
                    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->a.kmod[i]),
                            (i_val & nModifiers[i]));
                }
                beryl_setting_value_get_keysym(value,&i_val);
                s=XKeysymToString(i_val);
                if (!s)
                {
                    s="None";
                }
                gtk_entry_set_text(GTK_ENTRY(widgets->a.key),s);
            }
            if (beryl_setting_get_can_set_edgemask(setting,&b_val) && b_val)
            {
                gint i;
                beryl_setting_value_get_edgemask(value,&i_val);
                for (i=0;i<SCREEN_EDGE_NUM;i++)
                {
                    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->a.e[i])
                            ,(i_val&(1<<i))?TRUE:FALSE);
                }
            }
            if (beryl_setting_get_can_set_bell(setting,&b_val) && b_val)
            {
                beryl_setting_value_get_bell(value,&b_val);
                gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->a.b),b_val);
            }
            break;
        case BERYL_SETTING_TYPE_LIST:
            {
                BerylSettingType typ;
                beryl_setting_get_list_type(setting,&typ);
                if (typ==BERYL_SETTING_TYPE_STRING &&
                        beryl_setting_get_allowed_strings(setting,&l_val)
                        && l_val)
                {
                    gint i=0;
                    for (;l_val;l_val=l_val->next)
                    {
                        GSList * list;
                        beryl_setting_value_get_value_list(value,&list);
                        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->l[i]),FALSE);
                        for (;list;list=list->next)
                        {
                            beryl_setting_value_get_string(list->data,&s_val);
                            if (strcmp(s_val,l_val->data)==0)
                                gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets->l[i]),TRUE);
                        }
                        i++;
                    }
                }
                else if (typ==BERYL_SETTING_TYPE_STRING)
                {
                    if (!doing_active_plugins(setting,TRUE))
                    {
                        gtk_list_store_clear(widgets->s.backing);
                        beryl_setting_value_get_value_list(value,&l_val);
                        for (;l_val;l_val=l_val->next)
                        {
                            GtkTreeIter it;
                            beryl_setting_value_get_string(l_val->data,&s_val);
                            gtk_list_store_append(widgets->s.backing,&it);
                            gtk_list_store_set(widgets->s.backing,&it,0,s_val,-1);
                        }
                    }
                }
            }
        default:
            break;
    }
    suspend_signals=FALSE;
}
void reset_option(GtkWidget * but, SETResetOption * ro)
{
    BerylSetting * o = ro->o;
    if (o->type==BERYL_SETTING_TYPE_BINDING)
    {
        switch(ro->resetType)
        {
            case 0:
                o->value.value.as_binding.button = o->default_value.value.as_binding.button;
                o->value.value.as_binding.button_mod_mask = o->default_value.value.as_binding.button_mod_mask;
                o->value.value.as_binding.enabled.value.button = o->default_value.value.as_binding.enabled.value.button;
                break;
            case 1:
                o->value.value.as_binding.keysym = o->default_value.value.as_binding.keysym;
                o->value.value.as_binding.key_mod_mask = o->default_value.value.as_binding.key_mod_mask;
                o->value.value.as_binding.enabled.value.key = o->default_value.value.as_binding.enabled.value.key;
                break;
            case 2:
                o->value.value.as_binding.edge_mask = o->default_value.value.as_binding.edge_mask;
                break;
            case 3:
                o->value.value.as_binding.on_bell = o->default_value.value.as_binding.on_bell;
                break;
        }
    }
    else
        beryl_setting_reset_to_default(o);
    beryl_settings_context_write(context);
    set_widgets_for_setting(o);
    beryl_settings_send_reload_signal();
}
gboolean has_hint(BerylSetting * o, const gchar * hint)
{
    GKeyFile * f = g_key_file_new();
    g_key_file_load_from_file(f,DATADIR "/beryl/hints",0,NULL);
    gchar * group = o->parent->name;
    if (!group) group="_";
    gchar * search_for=g_strdup_printf(o->is_screen?"h_s_%s":"h_a_%s",o->name);
    gsize nHints;
    gchar ** hints = g_key_file_get_string_list(f,group,search_for,&nHints,NULL);
    g_free(search_for);
    g_key_file_free(f);
    gint i;
    for (i=0;i<nHints && hints && hints[i];i++)
    {
        if (strcmp(hints[i],hint)==0)
        {
            g_strfreev(hints);
            return TRUE;
        }
    }
    g_strfreev(hints);
    return FALSE;
}

    static void
cb_preview (GtkFileChooser *file_chooser, gpointer data)
{
    GtkWidget *preview;
    char *filename;
    GdkPixbuf *pixbuf;
    gboolean have_preview;

    preview = GTK_WIDGET (data);
    filename = gtk_file_chooser_get_preview_filename (file_chooser);
    pixbuf = NULL;
    have_preview = FALSE;
    if (filename)
    {
        pixbuf = gdk_pixbuf_new_from_file_at_size (filename, 128, 128, NULL);
        have_preview = (pixbuf != NULL);
        g_free (filename);

        if (pixbuf)
            gtk_image_set_from_pixbuf (GTK_IMAGE (preview), pixbuf);
    }
    if (pixbuf)
        g_object_unref (pixbuf);

    gtk_file_chooser_set_preview_widget_active (file_chooser, have_preview);
}
void modify_sl(BerylSetting * o)
{
    //re-write the o->nValue/o->value.l based on o->widgets.s.backing
    gint i;
    GtkTreeIter it;
    SETOptionWidgets * widgets = o->private_ptr;
    if (suspend_signals)
        return;
    beryl_setting_list_clear(o);
    //now we have to figure out how many rows...
    gint nValue=gtk_tree_model_iter_n_children(GTK_TREE_MODEL(widgets->s.backing),NULL);
    if (nValue)
    {
        for (i=0;i<nValue;i++)
        {
            const gchar * s;
            gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(widgets->s.backing),&it,NULL,i);
            gtk_tree_model_get(GTK_TREE_MODEL(widgets->s.backing),&it,0,&s,-1);
            beryl_setting_value_set_string(beryl_setting_list_append(o),&s);
        }
    }
    beryl_settings_context_write(context);
    beryl_settings_send_reload_signal();
}
void cb_browse(GtkWidget * w, BerylSetting * o)
{
    //make up a file widget
    GtkWidget * fd;
    gchar * titl=g_strconcat(_("Browse for "),beryl_setting_get_short_desc(o),NULL);
    fd = gtk_file_chooser_dialog_new(titl,GTK_WINDOW(mainWindow),
            has_hint(o,"directory")?GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER:
            GTK_FILE_CHOOSER_ACTION_OPEN,
            GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
            GTK_STOCK_OPEN,GTK_RESPONSE_ACCEPT,NULL);
    g_free(titl);

    if (has_hint(o,"image"))
    {
        GtkFileFilter * f;
        GtkWidget * pr;
        f = gtk_file_filter_new();
        if (has_hint(o,"pngonly"))
        {
            gtk_file_filter_set_name(f,"PNG Images");
        }
        else
        {
            gtk_file_filter_set_name(f,_("Beryl Image Types (PNG,SVG)"));
            gtk_file_filter_add_mime_type(f,"image/svg");
            gtk_file_filter_add_pattern(f,"*.svg");
            gtk_file_filter_add_pattern(f,"*.SVG");
        }
        gtk_file_filter_add_mime_type(f,"image/png");
        gtk_file_filter_add_pattern(f,"*.png");
        gtk_file_filter_add_pattern(f,"*.PNG");
        gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(fd),f);
        gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(fd),f);
        pr = gtk_image_new();
        gtk_file_chooser_set_preview_widget(GTK_FILE_CHOOSER(fd),pr);
        g_signal_connect(fd,_("update-preview"),G_CALLBACK(cb_preview),pr);
    }
    SETOptionWidgets * widgets = o->private_ptr;
    if (o->type == BERYL_SETTING_TYPE_STRING)
        if (g_file_test(gtk_entry_get_text(widgets->sb.e),G_FILE_TEST_EXISTS))
            gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(fd),gtk_entry_get_text(widgets->sb.e));
    if (gtk_dialog_run(GTK_DIALOG(fd))==GTK_RESPONSE_ACCEPT)
    {
        gchar * s;
        s=gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fd));
        if (o->type == BERYL_SETTING_TYPE_STRING)
            gtk_entry_set_text(widgets->sb.e,s);
        else
        {
            GtkTreeIter it;
            GtkTreeSelection * se = gtk_tree_view_get_selection(widgets->s.view);
            if (suspend_signals)
                return;
            gtk_tree_selection_unselect_all(se);
            gtk_list_store_append(widgets->s.backing,&it);
            gtk_list_store_set(widgets->s.backing,&it,0,s,-1);
            gtk_tree_selection_select_iter(se,&it);
            modify_sl(o);
        }
        g_free(s);
    }
    gtk_widget_destroy(fd);
}
GtkWidget * make_info_button_full(BerylSetting * o,gint rt)
{
    GtkWidget * but;
    GtkWidget * img;
    GtkWidget * hb;
    SETResetOption * ro = malloc(sizeof(SETResetOption));
    gboolean str=((o->type==BERYL_SETTING_TYPE_BINDING && rt!=3) || o->type==BERYL_SETTING_TYPE_LIST);
    if (str)
        hb = gtk_vbox_new(FALSE,2);
    else
        hb = gtk_hbox_new(TRUE,2);
    but = gtk_button_new();
    img = gtk_image_new_from_stock(GTK_STOCK_INFO,GTK_ICON_SIZE_BUTTON);
    gtk_container_add(GTK_CONTAINER(but),img);
    g_signal_connect(but,"clicked",G_CALLBACK(set_info_pane),getmark(o));
    gtk_box_pack_start(GTK_BOX(hb),but,str,str,0);
    but = gtk_button_new();
    img = gtk_image_new_from_stock(GTK_STOCK_CLEAR,GTK_ICON_SIZE_BUTTON);
    gtk_container_add(GTK_CONTAINER(but),img);
    ro->o=o;
    ro->resetType=rt;
    g_signal_connect(but,"clicked",G_CALLBACK(reset_option),ro);
    gtk_box_pack_start(GTK_BOX(hb),but,str,str,0);
    return hb;
}
GtkWidget * make_info_button(BerylSetting * o)
{
    return make_info_button_full(o,0);
}
void do_if_toggle_group(BerylSetting * o, SETGroupWidget * w)
{
    if (o->type==BERYL_SETTING_TYPE_BOOL)
    {
        GtkWidget * cbox;
        GtkWidget * box;
        SETOptionWidgets * widgets = o->private_ptr;
        box = gtk_hbox_new(FALSE,2);
        cbox = gtk_check_button_new_with_label(beryl_setting_get_short_desc(o));
        gtk_box_pack_start(GTK_BOX(w->widget),box,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(box),make_info_button(o),FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(box),cbox,TRUE,TRUE,0);
        widgets->w=cbox;
        g_signal_connect(cbox,"toggled",G_CALLBACK(cb_option),o);
        set_widgets_for_setting(o);
        w->hasany=TRUE;
    }
}
void do_if_numeric_group(BerylSetting * o, SETGroupWidget * w)
{
    if ((o->type==BERYL_SETTING_TYPE_INT || o->type==BERYL_SETTING_TYPE_FLOAT))
    {
        GtkWidget * slid;
        GtkWidget * spun;
        GtkWidget * box;
        GtkWidget * slidbox;
        SETOptionWidgets * widgets = o->private_ptr;
        gdouble min = o->type==BERYL_SETTING_TYPE_INT?o->info.for_int.min:o->info.for_float.min;
        gdouble max = o->type==BERYL_SETTING_TYPE_INT?o->info.for_int.max:o->info.for_float.max;
        gdouble pre = o->type==BERYL_SETTING_TYPE_INT?1:o->info.for_float.precision;
        slid = gtk_hscale_new_with_range(min,max,pre);
        spun = gtk_spin_button_new_with_range(min,max,pre);
        box = gtk_hbox_new(FALSE,2);
        slidbox = gtk_vbox_new(FALSE,2);
        gtk_box_pack_start(GTK_BOX(w->widget),box,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(box),make_info_button(o),FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(box),slidbox,TRUE,TRUE,0);
        gtk_box_pack_start(GTK_BOX(box),spun,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(slidbox),my_label_new(beryl_setting_get_short_desc(o)),FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(slidbox),slid,FALSE,FALSE,0);
        widgets->n.h=GTK_HSCALE(slid);
        widgets->n.s=GTK_SPIN_BUTTON(spun);
        gtk_scale_set_draw_value(GTK_SCALE(slid),FALSE);
        gtk_range_set_update_policy(GTK_RANGE(slid),GTK_UPDATE_DISCONTINUOUS);
        gtk_spin_button_set_update_policy(GTK_SPIN_BUTTON(spun),GTK_UPDATE_IF_VALID);
        g_signal_connect(slid,"value-changed",G_CALLBACK(cb_option),o);
        g_signal_connect(spun,"value-changed",G_CALLBACK(cb_option),o);
        set_widgets_for_setting(o);
        w->hasany=TRUE;
    }
}
void do_if_string_choice_group(BerylSetting * o, SETGroupWidget * w)
{
    if ((o->type==BERYL_SETTING_TYPE_STRING) && o->info.for_string.allowed_values)
    {
        GtkWidget * hbox = gtk_hbox_new(FALSE,2);
        GtkWidget * vbox = gtk_vbox_new(FALSE,2);
        GSList * i;
        SETOptionWidgets * widgets = o->private_ptr;
        gtk_box_pack_start(GTK_BOX(w->widget),hbox,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),make_info_button(o),FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),vbox,TRUE,TRUE,0);
        gtk_box_pack_start(GTK_BOX(vbox),my_label_new(beryl_setting_get_short_desc(o)),FALSE,FALSE,0);
        widgets->w=gtk_combo_box_new_text();
        for (i=o->info.for_string.allowed_values;i;i=i->next)
        {
            gtk_combo_box_append_text(GTK_COMBO_BOX(widgets->w),i->data); // yes this is correct
        }
        set_widgets_for_setting(o);
        g_signal_connect(widgets->w,"changed",G_CALLBACK(cb_option),o);
        gtk_box_pack_start(GTK_BOX(vbox),widgets->w,FALSE,FALSE,0);
        w->hasany=TRUE;
    }
}

void do_if_string_group(BerylSetting * o, SETGroupWidget * w)
{
    if ((o->type==BERYL_SETTING_TYPE_STRING) && !o->info.for_string.allowed_values)
    {
        GtkWidget * hbox = gtk_hbox_new(FALSE,2);
        GtkWidget * vbox = gtk_vbox_new(FALSE,2);
        GtkWidget * hbox2 = gtk_hbox_new(FALSE,2);
        SETOptionWidgets * widgets = o->private_ptr;
        gtk_box_pack_start(GTK_BOX(w->widget),hbox,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),make_info_button(o),FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),vbox,TRUE,TRUE,0);
        gtk_box_pack_start(GTK_BOX(vbox),my_label_new(beryl_setting_get_short_desc(o)),FALSE,FALSE,0);
        widgets->w = gtk_entry_new();
        set_widgets_for_setting(o);
        g_signal_connect(widgets->w,"changed",G_CALLBACK(cb_option),o);
        gtk_box_pack_start(GTK_BOX(hbox2),widgets->w,TRUE,TRUE,0);
        if (has_hint(o,"file"))
        {
            widgets->sb.b=GTK_BUTTON(gtk_button_new_with_label(_("Browse")));
            gtk_button_set_image(widgets->sb.b,gtk_image_new_from_stock(GTK_STOCK_OPEN,GTK_ICON_SIZE_BUTTON));
            gtk_box_pack_start(GTK_BOX(hbox2),GTK_WIDGET(widgets->sb.b),FALSE,FALSE,0);
            g_signal_connect(widgets->sb.b,"clicked",G_CALLBACK(cb_browse),o);
        }
        gtk_box_pack_start(GTK_BOX(vbox),hbox2,FALSE,FALSE,0);
        w->hasany=TRUE;
    }
}
void do_if_color_group(BerylSetting * o, SETGroupWidget * w)
{
    if ((o->type==BERYL_SETTING_TYPE_COLOR))
    {
        GtkWidget * hbox;
        GtkWidget * cbut;
        SETOptionWidgets * widgets = o->private_ptr;
        hbox = gtk_hbox_new(FALSE,2);
        gtk_box_pack_start(GTK_BOX(w->widget),hbox,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),make_info_button(o),FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),my_label_new(beryl_setting_get_short_desc(o)),TRUE,TRUE,0);
        cbut=gtk_color_button_new();
        gtk_color_button_set_use_alpha(GTK_COLOR_BUTTON(cbut),TRUE);
        gtk_box_pack_start(GTK_BOX(hbox),cbut,FALSE,FALSE,0);
        widgets->w=cbut;
        g_signal_connect(cbut,"color-set",G_CALLBACK(cb_option),o);
        set_widgets_for_setting(o);
        w->hasany=TRUE;
    }
}
void do_if_action_group(BerylSetting * o, SETGroupWidget * w)
{
    if (o->type==BERYL_SETTING_TYPE_BINDING)
    {
        w->hasany=TRUE;
        set_widgets_for_setting(o);
    }
}
void do_if_bell_group(BerylSetting * o, SETGroupWidget * w)
{
    if ((o->type==BERYL_SETTING_TYPE_BINDING) && o->info.for_bind.bell)
    {
        GtkWidget * hbox = gtk_hbox_new(FALSE,2);
        SETOptionWidgets * widgets = o->private_ptr;
        gtk_box_pack_start(GTK_BOX(w->widget),hbox,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),make_info_button_full(o,3),FALSE,FALSE,0);
        widgets->a.b=GTK_CHECK_BUTTON(gtk_check_button_new_with_label(
                    g_strconcat(beryl_setting_get_short_desc(o)," on System Bell",NULL)));
        gtk_box_pack_start(GTK_BOX(hbox),GTK_WIDGET(widgets->a.b),TRUE,TRUE,0);
        g_signal_connect(widgets->a.b,"toggled",G_CALLBACK(cb_option),o);
        w->hasany=TRUE;
    }
}
void cb_getkey(GtkWidget * w, BerylSetting * o)
{
    // shamelessly stolen from e16keyedit & modified slightly
    GtkWidget *win, *label, *frame, *align;

    win = gtk_window_new(GTK_WINDOW_POPUP);
    gtk_window_set_policy(GTK_WINDOW(win), 0, 0, 1);
    gtk_window_set_position(GTK_WINDOW(win), GTK_WIN_POS_MOUSE);
    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);
    align = gtk_alignment_new(0.0, 0.0, 0.0, 0.0);
    gtk_container_set_border_width(GTK_CONTAINER(align), 32);
    label = gtk_label_new(_("Please press the key on the keyboard\n"
                "you wish to modify this keyboard-shortcut\n"
                "to use from now on. (Do not worry about modifiers)\n"
                "Press ESCAPE to cancel."));
    gtk_container_add(GTK_CONTAINER(win), frame);
    gtk_container_add(GTK_CONTAINER(frame), align);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_widget_show_all(win);
    while (gtk_events_pending())
        gtk_main_iteration();
    gdk_flush();
    while (gtk_events_pending())
        gtk_main_iteration();
    {
        char *key;
        XEvent ev;

        gdk_window_set_events(win->window, GDK_KEY_PRESS_MASK);
        XSetInputFocus(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(win->window),
                RevertToPointerRoot, CurrentTime);
        gdk_keyboard_grab(win->window, TRUE, CurrentTime);
        XWindowEvent(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(win->window),
                KeyPressMask, &ev);
        gdk_keyboard_ungrab(GDK_CURRENT_TIME);
        key = XKeysymToString(XKeycodeToKeysym(GDK_DISPLAY(),
                    ev.xkey.keycode, 0));
        SETOptionWidgets * widgets = o->private_ptr;
        if (strcmp(key,"Escape")!=0)
            gtk_entry_set_text(GTK_ENTRY(widgets->a.key),key);
    }
    gtk_widget_destroy(win);
}
void make_key_button_stuff(BerylSetting * o, SETGroupWidget * w, gint p)
{
    GtkButton * captbut;
    GtkEntry * entry;
    GtkCheckButton ** mods; 
    gint i;
    GtkWidget * tab;
    GtkWidget * vbox2;
    GtkWidget * hbox2;
    GtkWidget * hbox;
    GtkWidget * fram;
    GtkWidget * ench;
    SETOptionWidgets * widgets=o->private_ptr;
    if (p==1)
        mods=widgets->a.kmod;
    else
        mods=widgets->a.bmod;
    fram = gtk_frame_new(beryl_setting_get_short_desc(o));
    gtk_box_pack_start(GTK_BOX(w->widget),fram,FALSE,FALSE,0);
    hbox = gtk_hbox_new(FALSE,2);
    gtk_container_add(GTK_CONTAINER(fram),hbox);
    gtk_box_pack_start(GTK_BOX(hbox),make_info_button_full(o,p),FALSE,FALSE,0);
    vbox2 = gtk_vbox_new(FALSE,2);
    gtk_box_pack_start(GTK_BOX(hbox),vbox2,TRUE,TRUE,0);
    ench=gtk_check_button_new_with_label("Enabled");
    gtk_box_pack_start(GTK_BOX(vbox2),ench,FALSE,FALSE,0);
    entry = GTK_ENTRY(gtk_entry_new());
    if (p==1)
    {
        widgets->a.key=entry;
        widgets->a.keyen=GTK_CHECK_BUTTON(ench);
    }
    else
    {
        widgets->a.but=entry;
        widgets->a.buten=GTK_CHECK_BUTTON(ench);
    }
    g_signal_connect(ench,"toggled",G_CALLBACK(cb_option),o);
    gtk_entry_set_width_chars(entry,8);
    hbox2=gtk_hbox_new(FALSE,2);
    gtk_box_pack_start(GTK_BOX(vbox2),hbox2,FALSE,FALSE,0);
    tab = gtk_table_new(0,0,FALSE);
    gtk_box_pack_start(GTK_BOX(hbox2),my_label_new(_("Modifier Keys:")),FALSE,FALSE,0);
    gtk_box_pack_start(GTK_BOX(hbox2),tab,TRUE,TRUE,0);
    for(i=0;i<NMODIFIERS;i++)
    {
        gint col = (i%(NMODIFIERS/MODROWS));
        gint row = (i/(NMODIFIERS/MODROWS));
        mods[i]=GTK_CHECK_BUTTON(gtk_check_button_new_with_label(_(modifiers[i])));
        g_signal_connect(mods[i],"toggled",G_CALLBACK(cb_option),o);
        gtk_table_attach_defaults(GTK_TABLE(tab),GTK_WIDGET(mods[i]),
                col,col+1,row,row+1);
    }
    hbox2=gtk_hbox_new(FALSE,2);
    gtk_box_pack_start(GTK_BOX(vbox2),hbox2,FALSE,FALSE,0);
    gtk_box_pack_start(GTK_BOX(hbox2),gtk_label_new(p==1?_("Key Name:"):_("Button Number:")),FALSE,FALSE,0);
    gtk_box_pack_start(GTK_BOX(hbox2),GTK_WIDGET(entry),TRUE,TRUE,0);
    if (p==1)
    {
        captbut=GTK_BUTTON(gtk_button_new_with_label(_("Grab Key")));
        g_signal_connect(captbut,"clicked",G_CALLBACK(cb_getkey),o);
        gtk_box_pack_start(GTK_BOX(hbox2),GTK_WIDGET(captbut),FALSE,FALSE,0);
    }
    g_signal_connect(entry,"changed",G_CALLBACK(cb_option),o);
}
void do_if_key_group(BerylSetting * o, SETGroupWidget * w)
{
    if (o->type==BERYL_SETTING_TYPE_BINDING && o->info.for_bind.key)
    {
        make_key_button_stuff(o,w,1);
        w->hasany=TRUE;
    }
}
void do_if_button_group(BerylSetting * o, SETGroupWidget * w)
{
    if (o->type==BERYL_SETTING_TYPE_BINDING && o->info.for_bind.button)
    {
        make_key_button_stuff(o,w,0);
        w->hasany=TRUE;
    }
}
void do_if_edge_group(BerylSetting * o, SETGroupWidget * w)
{
    if (o->type==BERYL_SETTING_TYPE_BINDING && o->info.for_bind.edge)
    {
        gint i;
        GtkWidget * fram2;
        GtkWidget * tab;
        GtkWidget * hbox;
        SETOptionWidgets * widgets = o->private_ptr;
        fram2=gtk_frame_new(beryl_setting_get_short_desc(o));
        tab=gtk_table_new(0,0,FALSE);
        gtk_box_pack_start(GTK_BOX(w->widget),fram2,FALSE,FALSE,0);
        hbox = gtk_hbox_new(FALSE,2);
        gtk_container_add(GTK_CONTAINER(fram2),hbox);
        gtk_box_pack_start(GTK_BOX(hbox),make_info_button_full(o,2),FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),tab,TRUE,TRUE,0);

        for (i=0;i<SCREEN_EDGE_NUM;i++)
        {
            widgets->a.e[i]=GTK_CHECK_BUTTON(gtk_check_button_new_with_label(_(edges[i])));
            gtk_table_attach_defaults(GTK_TABLE(tab),GTK_WIDGET(widgets->a.e[i]),edgecols[i],edgecols[i]+1,edgerows[i],edgerows[i]+1);
            g_signal_connect(widgets->a.e[i],"toggled",G_CALLBACK(cb_option),o);
        }
        w->hasany=TRUE;
    }
}
#define OPT_PER_LINE 3
void do_if_option_list_group(BerylSetting * o, SETGroupWidget * w)
{
    if (o->type==BERYL_SETTING_TYPE_LIST && o->info.for_list.list_of_type==BERYL_SETTING_TYPE_STRING &&
            o->info.for_list.list_of_info->for_string.allowed_values)
    {
        GtkWidget * hbox, *vbox, *tabl, *fram;
        gint i, l=-1, c=0;
        GSList * j;
        hbox = gtk_hbox_new(FALSE,2);
        SETOptionWidgets * widgets = o->private_ptr;
        fram = gtk_frame_new(beryl_setting_get_short_desc(o));
        gtk_box_pack_start(GTK_BOX(w->widget),fram,FALSE,FALSE,0);
        gtk_container_add(GTK_CONTAINER(fram),hbox);
        gtk_box_pack_start(GTK_BOX(hbox),make_info_button(o),FALSE,FALSE,0);
        vbox = gtk_vbox_new(FALSE,2);
        gtk_box_pack_start(GTK_BOX(hbox),vbox,TRUE,TRUE,0);
        tabl=gtk_table_new(OPT_PER_LINE,1,TRUE);
        gtk_box_pack_start(GTK_BOX(vbox),tabl,FALSE,FALSE,0);
        i=0;
        widgets->l=malloc(sizeof(GtkWidget *)*g_slist_length(o->info.for_list.list_of_info->for_string.allowed_values));
        for (j=o->info.for_list.list_of_info->for_string.allowed_values;j;j=j->next)
        {
            GtkWidget * cbut;
            if (i%OPT_PER_LINE==0)
            {
                l++;
                c=0;
            }
            else
                c++;
            cbut = gtk_check_button_new_with_label(j->data);
            gtk_table_attach_defaults(GTK_TABLE(tabl),cbut,c,c+1,l,l+1);
            widgets->l[i]=cbut;
            g_signal_connect(cbut,"toggled",G_CALLBACK(cb_option),o);
            i++;
        }
        set_widgets_for_setting(o);
        w->hasany=TRUE;
    }
}
void cb_sl_edited(GtkWidget * w, gchar * treepath, gchar * value, BerylSetting * o)
{
    GtkTreeIter it;
    if (suspend_signals)
        return;
    SETOptionWidgets * widgets = o->private_ptr;
    gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(widgets->s.backing),&it,treepath);
    gtk_list_store_set(widgets->s.backing,&it,0,value,-1);
    //modify_sl(o);
}
void cb_sl_select(GtkTreeSelection * sel, BerylSetting * o)
{
    GtkTreeIter it;
    GtkTreeModel * mod;
    gboolean p;
    SETOptionWidgets * widgets=o->private_ptr;
    p=gtk_tree_selection_get_selected(sel,&mod,&it);
    gtk_widget_set_sensitive(GTK_WIDGET(widgets->s.rem),p);
    gtk_widget_set_sensitive(GTK_WIDGET(widgets->s.up),p);
    gtk_widget_set_sensitive(GTK_WIDGET(widgets->s.dn),p);
}
void cb_sl_add(GtkWidget * w, BerylSetting * o)
{
    GtkTreeIter it;
    SETOptionWidgets * widgets=o->private_ptr;
    GtkTreeSelection * s = gtk_tree_view_get_selection(widgets->s.view);
    if (suspend_signals)
        return;
    gtk_tree_selection_unselect_all(s);
    gtk_list_store_append(widgets->s.backing,&it);
    gtk_list_store_set(widgets->s.backing,&it,0,"<New Value>",-1);
    gtk_tree_selection_select_iter(s,&it);
}
void cb_sl_rem(GtkWidget * w, BerylSetting * o)
{
    GtkTreeIter it;
    GtkTreeModel * mod;
    SETOptionWidgets * widgets=o->private_ptr;
    GtkTreeSelection * s = gtk_tree_view_get_selection(widgets->s.view);
    if (suspend_signals)
        return;
    if (gtk_tree_selection_get_selected(s,&mod,&it))
        gtk_list_store_remove(GTK_LIST_STORE(mod),&it);
}
void cb_row_changed(GtkWidget * w, GtkTreePath * p, GtkTreeIter * i, BerylSetting *o)
{
    if (suspend_signals)
        return;
    modify_sl(o);
}
void cb_row_del(GtkWidget * w, GtkTreePath * p, BerylSetting *o)
{
    if (suspend_signals)
        return;
    modify_sl(o);
}
void move_item(BerylSetting * o, gboolean up)
{
    GtkTreeIter it;
    GtkTreeModel * mod;
    SETOptionWidgets * widgets=o->private_ptr;
    GtkTreeSelection * s = gtk_tree_view_get_selection(widgets->s.view);
    if (suspend_signals)
        return;
    if (gtk_tree_selection_get_selected(s,&mod,&it))
    {
        GtkTreeIter nit;
        if (up)
        {
            gchar * s = gtk_tree_model_get_string_from_iter(mod,&it);
            GtkTreePath * p = gtk_tree_path_new_from_string(s);
            g_free(s);
            if (!gtk_tree_path_prev(p))
            {
                g_free(p);
                return;
            }
            s=gtk_tree_path_to_string(p);
            gtk_tree_model_get_iter_from_string(mod,&nit,s);
            g_free(s);
        }
        else
        {
            nit=it;
            if (!gtk_tree_model_iter_next(mod,&nit))
                return;
        }
        gtk_list_store_swap(GTK_LIST_STORE(mod),&it,&nit);
    }
}
void cb_move_up(GtkWidget * w, BerylSetting * o)
{
    move_item(o,TRUE);
}
void cb_move_dn(GtkWidget * w, BerylSetting * o)
{
    move_item(o,FALSE);
}
void do_if_string_list_group(BerylSetting * o, SETGroupWidget * w)
{
    if (o->type==BERYL_SETTING_TYPE_LIST && o->info.for_list.list_of_type==BERYL_SETTING_TYPE_STRING &&
            !o->info.for_list.list_of_info->for_string.allowed_values)
    {
        GtkWidget * hbox;
        GtkCellRenderer * celr;
        GtkWidget * fram;
        GtkTreeViewColumn * col;
        GtkScrolledWindow * win;
        GtkWidget * vbox;
        SETOptionWidgets * widgets = o->private_ptr;
        if (!o->parent->name && strcmp(o->name,"a_active_plugins")==0)
        {
            return;
        }
        fram = gtk_frame_new(beryl_setting_get_short_desc(o));
        hbox = gtk_hbox_new(FALSE,2);
        gtk_box_pack_start(GTK_BOX(w->widget),fram,FALSE,FALSE,0);
        gtk_container_add(GTK_CONTAINER(fram),hbox);
        gtk_box_pack_start(GTK_BOX(hbox),make_info_button(o),FALSE,FALSE,0);
        widgets->s.backing=gtk_list_store_new(1,G_TYPE_STRING);
        celr = GTK_CELL_RENDERER(gtk_cell_renderer_text_new());
        g_signal_connect(celr,"edited",G_CALLBACK(cb_sl_edited),o);
        g_object_set(celr,"editable",TRUE,NULL);
        col = gtk_tree_view_column_new_with_attributes(beryl_setting_get_short_desc(o),GTK_CELL_RENDERER(celr),"text",0,NULL);
        widgets->s.view=GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(widgets->s.backing)));
        gtk_tree_view_set_reorderable(widgets->s.view,TRUE);
        gtk_tree_view_set_headers_visible(widgets->s.view,FALSE);
        gtk_tree_view_append_column(widgets->s.view,col);
        win = GTK_SCROLLED_WINDOW(gtk_scrolled_window_new(NULL,NULL));
        gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win),GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
        gtk_container_add(GTK_CONTAINER(win),GTK_WIDGET(widgets->s.view));
        gtk_box_pack_start(GTK_BOX(hbox),GTK_WIDGET(win),TRUE,TRUE,0);
        widgets->s.add=GTK_BUTTON(gtk_button_new_from_stock(GTK_STOCK_ADD));
        if (has_hint(o,"file"))
        {
            widgets->s.bro=GTK_BUTTON(gtk_button_new_with_label(_("Browse")));
            gtk_button_set_image(widgets->s.bro,gtk_image_new_from_stock(
                        GTK_STOCK_OPEN,GTK_ICON_SIZE_BUTTON));
            g_signal_connect(widgets->s.bro,"clicked",G_CALLBACK(cb_browse),o);
        }
        widgets->s.rem=GTK_BUTTON(gtk_button_new_from_stock(GTK_STOCK_DELETE));
        vbox = gtk_vbox_new(FALSE,2);
        gtk_box_pack_start(GTK_BOX(hbox),vbox,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(vbox),GTK_WIDGET(
                    widgets->s.up=GTK_BUTTON(gtk_button_new())),FALSE,FALSE,0);
        gtk_box_pack_end(GTK_BOX(vbox),GTK_WIDGET(
                    widgets->s.dn=GTK_BUTTON(gtk_button_new())),FALSE,FALSE,0);
        gtk_container_add(GTK_CONTAINER(widgets->s.up),
                gtk_image_new_from_stock(GTK_STOCK_GO_UP,GTK_ICON_SIZE_BUTTON));
        gtk_container_add(GTK_CONTAINER(widgets->s.dn),
                gtk_image_new_from_stock(GTK_STOCK_GO_DOWN,GTK_ICON_SIZE_BUTTON));
        gtk_widget_set_sensitive(GTK_WIDGET(widgets->s.rem),FALSE);
        gtk_widget_set_sensitive(GTK_WIDGET(widgets->s.up),FALSE);
        gtk_widget_set_sensitive(GTK_WIDGET(widgets->s.dn),FALSE);
        vbox = gtk_vbox_new(FALSE,2);
        gtk_box_pack_start(GTK_BOX(hbox),vbox,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(vbox),GTK_WIDGET(widgets->s.add),FALSE,FALSE,0);
        if (has_hint(o,"file"))
            gtk_box_pack_start(GTK_BOX(vbox),GTK_WIDGET(widgets->s.bro),FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(vbox),GTK_WIDGET(widgets->s.rem),FALSE,FALSE,0);
        g_signal_connect(gtk_tree_view_get_selection(widgets->s.view),"changed",
                G_CALLBACK(cb_sl_select),o);
        g_signal_connect(widgets->s.add,"clicked",G_CALLBACK(cb_sl_add),o);
        g_signal_connect(widgets->s.rem,"clicked",G_CALLBACK(cb_sl_rem),o);
        g_signal_connect(widgets->s.up,"clicked",G_CALLBACK(cb_move_up),o);
        g_signal_connect(widgets->s.dn,"clicked",G_CALLBACK(cb_move_dn),o);
        gtk_tree_view_set_rules_hint(widgets->s.view,TRUE);
        set_widgets_for_setting(o);
        g_signal_connect(widgets->s.backing,"row-changed",G_CALLBACK(cb_row_changed),o);
        g_signal_connect(widgets->s.backing,"row-inserted",G_CALLBACK(cb_row_changed),o);
        g_signal_connect(widgets->s.backing,"row-deleted",G_CALLBACK(cb_row_del),o);
        w->hasany=TRUE;
    }
}
void do_if_command_string_group(BerylSetting * o, SETGroupWidget * w)
{
    if (has_hint(o,"command"))
        do_if_string_group(o,w);
}
void do_if_simple_string_group(BerylSetting * o, SETGroupWidget * w)
{
    if (!has_hint(o,"file") && !has_hint(o,"command"))
        do_if_string_group(o,w);
}
void do_if_file_string_group(BerylSetting * o, SETGroupWidget * w)
{
    if (has_hint(o,"file"))
    {
        do_if_string_group(o,w);
        do_if_string_list_group(o,w);
    }
}
void do_if_non_file_string_list_group(BerylSetting * o, SETGroupWidget * w)
{
    if (!has_hint(o,"file"))
        do_if_string_list_group(o,w);
}
void do_if_choice_option_group(BerylSetting * o, SETGroupWidget * w)
{
    do_if_toggle_group(o,w);
    do_if_string_choice_group(o,w);
    do_if_option_list_group(o,w);
    do_if_bell_group(o,w);
}
void setup_actions(gchar * group)
{
    SETGroupWidget z;
    z.group=group;
    z.widget=NULL;
    z.hasany=FALSE;
    if (strcmp(group,"_")==0) group=NULL;
    GSList * settings = beryl_settings_plugin_get_settings(beryl_settings_context_find_plugin(context,group));
    g_slist_foreach(settings,(GFunc)do_if_action_group,&z);
}
void do_page(gchar * group,GFunc func,gchar * label, GtkWidget * notebook)
{
    SETGroupWidget z;
    GtkWidget * scrol = gtk_scrolled_window_new(NULL,NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrol)
            ,GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
    GtkWidget * vbox = gtk_vbox_new(FALSE,6);
    gtk_container_set_border_width(GTK_CONTAINER(vbox),4);
    z.group=group;
    z.widget=vbox;
    z.hasany=FALSE;
    if (strcmp(group,"_")==0) group=NULL;
    GSList * settings = beryl_settings_plugin_get_settings(beryl_settings_context_find_plugin(context,group));
    g_slist_foreach(settings,func,&z);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrol),vbox);
    if (z.hasany)
        gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
                scrol,gtk_label_new(label));
}
void do_option_page(gchar * group, GtkWidget * notebook)
{
    //load_options_for(group);
    //okay now we can lay out a notebook
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook),TRUE);
    do_page(group,(GFunc)do_if_choice_option_group,_("Choices"),notebook);
    do_page(group,(GFunc)do_if_numeric_group,_("Numeric Values"),notebook);
    do_page(group,(GFunc)do_if_command_string_group,_("Commands"),notebook);
    do_page(group,(GFunc)do_if_simple_string_group,_("Strings"),notebook);
    do_page(group,(GFunc)do_if_file_string_group,_("Filenames"),notebook);
    do_page(group,(GFunc)do_if_color_group,_("Colors"),notebook);
    do_page(group,(GFunc)do_if_key_group,_("Keyboard"),notebook);
    do_page(group,(GFunc)do_if_button_group,_("Mouse"),notebook);
    do_page(group,(GFunc)do_if_edge_group,_("Screen Edges & Corners"),notebook);
    do_page(group,(GFunc)do_if_non_file_string_list_group,_("String Lists"),notebook);
    setup_actions(group);
    //gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
    //        do_page(group,(GFunc)do_if_undefined_group),gtk_label_new("Unhandled"));
}
gboolean get_plugin_active(gchar * plugin)
{
    GtkTreeIter it;
    gboolean k,b;
    if (strcmp(plugin,"settings")==0)
        return TRUE;
    b=FALSE;
    for(k=gtk_tree_model_get_iter_first(GTK_TREE_MODEL(pluginList),&it);k;
            k=gtk_tree_model_iter_next(GTK_TREE_MODEL(pluginList),&it))
    {
        gchar * pl;
        gtk_tree_model_get(GTK_TREE_MODEL(pluginList),&it,PluginName,&pl,-1);
        if (strcmp(pl,plugin)==0)
        {
            gtk_tree_model_get(GTK_TREE_MODEL(pluginList),&it,PluginEnabled,&b,-1);
        }
        g_free(pl);
    }
    return b;
}
void godoing_active_plugins(BerylSetting * o, gpointer p)
{
    doing_active_plugins(o,FALSE);
}
void do_active_plugins()
{
    GSList * settings=beryl_settings_plugin_get_settings(beryl_settings_context_find_plugin(context,NULL));
    g_slist_foreach(settings,(GFunc)godoing_active_plugins,NULL);
}
void cb_showpage(GtkWidget * w, GtkWidget * p)
{
    gint i;
    GtkTreeIter it;
    GtkTreeSelection * Select;
    Select=GTK_TREE_SELECTION(w);
    GtkTreeModel * mod;
    gchar * nam;
    if (gtk_bin_get_child(GTK_BIN(p)))
    {
        gtk_container_remove(GTK_CONTAINER(p),gtk_bin_get_child(GTK_BIN(p)));
    }
    if (gtk_tree_selection_get_selected(Select,&mod,&it))
    {
        gtk_tree_model_get(mod,&it,PluginName,&nam,-1);
    }
    else
        return;
    if (!nam)
        return;
    for (i=0;i<nPages;i++)
    {
        if (strcmp(SETOptionPages[i].name,nam)==0)
        {
            gtk_container_add(GTK_CONTAINER(p),SETOptionPages[i].widget);
            gtk_widget_show_all(SETOptionPages[i].widget);
            return;
        }
    }
    g_free(nam);
}
void beryl_settings_deactivate_plugin(gchar * plugin);
void beryl_settings_activate_plugin(gchar * plugin);

void check_deps_satisfied(SETPlugin * p)
{
    //check that all features are provide'd, and that all require's are loaded
    //if not, call beryl_settings_deactivate_plugin
    GSList * r;
    for (r=p->plugin->requires;r;r=r->next)
    {
        GSList * pl;
        gboolean found=FALSE;
        for (pl=Plugins;pl;pl=pl->next)
        {
            SETPlugin * q = pl->data;
            GSList * s;
            if (q->active)
            {
                for(s=q->plugin->provides;s;s=s->next)
                {
                    if (!g_strcasecmp(s->data,r->data))
                    {
                        found=TRUE;
                        break;
                    }
                }
            }
            if (found)
                break;
        }
        if (!found)
        {
            beryl_settings_deactivate_plugin(p->plugin->name);
            return;
        }
    }
}

void beryl_settings_deactivate_plugin(gchar * plugin)
{
    GSList * pl;
    set_plugin_active(plugin,FALSE);
    for (pl=Plugins;pl;pl=pl->next)
    {
        if (((SETPlugin*)pl->data)->active)
            check_deps_satisfied(pl->data);
    }
}
int query_which_activate(SETPlugin * p, gchar * feature, SETPlugin ** prov, int nprov)
{
    GtkWidget * w;
    GtkWidget * vbox;
    GtkWidget * combobox;
    GtkWidget * l;
    gchar * s;
    gint i;
    s=g_markup_printf_escaped(_("Plugin <i>%s</i> needs feature <i>%s</i>.\n"
                "This feature is provided by multiple plugins.\n"
                "Please select which plugin to activate\n"
                "to provide <i>%s</i> for <i>%s</i>"),p->plugin->short_desc?p->plugin->short_desc:p->plugin->name
            ,feature,feature,p->plugin->short_desc?p->plugin->short_desc:p->plugin->name);
    w=gtk_dialog_new_with_buttons(_("Beryl Settings - Resolve Dependancy"),
            GTK_WINDOW(mainWindow),GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_STOCK_OK,GTK_RESPONSE_ACCEPT,GTK_STOCK_CANCEL,GTK_RESPONSE_REJECT,NULL);
    vbox = GTK_DIALOG(w)->vbox;
    l=gtk_label_new("");
    gtk_label_set_markup(GTK_LABEL(l),s);
    gtk_box_pack_start(GTK_BOX(vbox),l,FALSE,FALSE,0);
    combobox=gtk_combo_box_new_text();
    for(i=0;i<nprov;i++)
        gtk_combo_box_append_text(
                GTK_COMBO_BOX(combobox),prov[i]->plugin->short_desc?prov[i]->plugin->short_desc:prov[i]->plugin->name);
    gtk_combo_box_set_active(GTK_COMBO_BOX(combobox),0);
    gtk_box_pack_start(GTK_BOX(vbox),combobox,FALSE,FALSE,0);
    gtk_widget_show_all(vbox);
    if (gtk_dialog_run(GTK_DIALOG(w))==GTK_RESPONSE_ACCEPT)
    {
        i=gtk_combo_box_get_active(GTK_COMBO_BOX(combobox));
        gtk_widget_destroy(w);
        return i;
    }
    else
    {
        gtk_widget_destroy(w);
    }
    return -1;
}
gboolean find_feature_provider(SETPlugin * p, gchar * feature)
{
    GSList * pl;
    int nprov=0;
    SETPlugin ** prov=NULL;
    for (pl=Plugins;pl;pl=pl->next)
    {
        if (check_contains(((SETPlugin *)pl->data)->plugin->provides,feature))
        {
            nprov++;
            prov=realloc(prov,sizeof(SETPlugin)*nprov);
            prov[nprov-1]=pl->data;
        }
    }
    if (nprov==1)
    {
        beryl_settings_activate_plugin(prov[0]->plugin->name);
        free(prov);
        return TRUE;
    }
    else if (nprov)
    {
        //here's where the magic happens
        int r = query_which_activate(p,feature,prov,nprov);
        if (r!=-1)
        {
            beryl_settings_activate_plugin(prov[r]->plugin->name);
        }
        free(prov);
        return (r!=-1);
    }
    else
        return FALSE;
}

gboolean fulfill_plugin_deps(SETPlugin * p)
{
    GSList * pl;
    GSList * s;
    for(s=p->plugin->requires;s;s=s->next)
    {
        gboolean found=FALSE;
        for(pl=Plugins;pl;pl=pl->next)
        {
            if (check_contains(((SETPlugin *)pl->data)->plugin->provides,s->data) && 
                    ((SETPlugin *)pl->data)->active)
            {
                found=TRUE;
                break;
            }
        }
        if (!found)
        {
            if (!find_feature_provider(p,s->data))
                return FALSE;
        }
    }
    return TRUE;
}
void clear_conflicts(SETPlugin * p)
{
    GSList * pl;
    for (pl=Plugins;pl;pl=pl->next)
    {
        SETPlugin * q = pl->data;
        if (q==p)
            continue;
        if (q->active && check_intersects(p->plugin->provides,q->plugin->provides))
            beryl_settings_deactivate_plugin(q->plugin->name);
    }
}
void beryl_settings_activate_plugin(gchar * plugin)
{
    //check plugin deps first
    SETPlugin * p=NULL;
    GSList * pl=NULL;
    set_plugin_active(plugin,TRUE); // makes other plugins' deps work right
    for(pl=Plugins;pl;pl=pl->next)
        if (!g_strcasecmp(((SETPlugin *)pl->data)->plugin->name?
                    ((SETPlugin *)pl->data)->plugin->name:"_",plugin))
            break;
    if (pl)
        p=pl->data;
    if (!p)
        return;
    clear_conflicts(p);
    if (!fulfill_plugin_deps(p))
        beryl_settings_deactivate_plugin(plugin);
}
void compile_plugin_order_list()
{
    SETPlugin * p;
    static gchar * prepend[]={"settings","crashhandler"};
    static const int nPrepend=sizeof(prepend)/sizeof(gchar *);
    gboolean preFound[]={FALSE};
    static gchar * append[]={"blurfx","bench"};
    static const int nAppend=sizeof(append)/sizeof(gchar *);
    gboolean appFound[]={FALSE,FALSE};
    Plugins=NULL;
    gint i;
    for (i=0;i<nPages;i++)
    {
        GSList * pl;
        gboolean found=FALSE;
        gint j;
        if (!strcmp("_",SETOptionPages[i].name))
            continue; // skip _ section
        for (j=0;j<nPrepend;j++)
            if(!strcmp(prepend[j],SETOptionPages[i].name))
            {
                preFound[j]=TRUE;
                found=TRUE;
                break;
            }
        for (j=0;j<nAppend;j++)
            if(!strcmp(append[j],SETOptionPages[i].name))
            {
                appFound[j]=TRUE;
                found=TRUE;
                break;
            }
        if (found)
            continue; // skip plugins with a fixed position
        p=malloc(sizeof(SETPlugin));
        memset(p,0,sizeof(SETPlugin));
        p->plugin=beryl_settings_context_find_plugin(context,SETOptionPages[i].name);
        p->active=FALSE;
        Plugins=g_slist_append(Plugins,p);
    }
    for (i=0;i<nAppend;i++)
    {
        if (!appFound[i])
            continue;
        p=malloc(sizeof(SETPlugin));
        memset(p,0,sizeof(SETPlugin));
        p->plugin=beryl_settings_context_find_plugin(context,append[i]);
        p->active=FALSE;
        Plugins=g_slist_append(Plugins,p);
    }
    for (i=nPrepend;i>0;i--)
    {
        if (!preFound[i-1])
            continue;
        p=malloc(sizeof(SETPlugin));
        memset(p,0,sizeof(SETPlugin));
        p->plugin=beryl_settings_context_find_plugin(context,prepend[i-1]);
        p->active=FALSE;
        Plugins=g_slist_prepend(Plugins,p);
    }
}
void cb_toggled(GtkCellRendererToggle * t, gchar * path, gpointer d)
{
    GtkTreeIter it;
    gboolean b;
    gchar * pn;
    GSList * pl;
    GtkTreePath * p = gtk_tree_path_new_from_string(path);
    if (!gtk_tree_model_get_iter(GTK_TREE_MODEL(pluginList),&it,p))
        g_error("invalid toggle");
    gtk_tree_path_free(p);
    gtk_tree_model_get(GTK_TREE_MODEL(pluginList),&it,PluginEnabled,&b,PluginName,&pn,-1);
    if (b)
        beryl_settings_deactivate_plugin(pn);
    else
        beryl_settings_activate_plugin(pn);
    g_free(pn);
    BerylSetting * act_plug;
    act_plug=beryl_settings_plugin_find_setting(
            beryl_settings_context_find_plugin(context,NULL),"active_plugins",FALSE);
    beryl_setting_list_clear(act_plug);
    for (pl=Plugins;pl;pl=pl->next)
    {
        if (((SETPlugin*)pl->data)->active)
        {
            BerylSettingValue * val = beryl_setting_list_append(act_plug);
            beryl_setting_value_set_string(val,(const gchar **)&((SETPlugin*)pl->data)->plugin->name);
        }
    }
    queue_option_changed();
}
void cb_main_destroy(GtkWidget * w, gpointer p)
{
    gtk_main_quit();
}
void set_widgets_for_plugin(BerylSettingsPlugin * plugin)
{
    g_slist_foreach(beryl_settings_plugin_get_settings(plugin),(GFunc)set_widgets_for_setting,NULL);
}
void reloadsig()
{
    beryl_settings_context_read(context);
    g_slist_foreach(beryl_settings_context_get_plugins(context),(GFunc)set_widgets_for_plugin,NULL);
    queue_option_changed();
}
void resettodefaults()
{
    gchar *resetdefbackup=g_strconcat(g_get_home_dir(),"/.beryl/settings.old",NULL);
    g_rename(profilefile,resetdefbackup);
    g_free(resetdefbackup);
}
void undoresettodefaults()
{
    gchar *resetdefbackup=g_strconcat(g_get_home_dir(),"/.beryl/settings.old",NULL);
    g_rename(resetdefbackup,profilefile);
    g_free(resetdefbackup);
}
void onResetChanged(GtkWidget *resetbutton, gpointer user_data)
{
    {
        gint idx=gtk_combo_box_get_active(GTK_COMBO_BOX(resetbutton));
        if (idx == 1)
            resettodefaults();
        else if (idx == 2)
            undoresettodefaults();
    }
    gtk_combo_box_set_active(GTK_COMBO_BOX(resetbutton),0);
    reloadsig();
}
void loadProfile (gint profileNumber)
{
    profilefile=g_strconcat(g_get_home_dir(),"/",profiles[profileNumber],NULL);
    gchar *settingsfile=g_strconcat(g_get_home_dir(),"/.beryl/settings",NULL);
    gchar *touchprofile=g_strconcat("touch ", profilefile,NULL);
    gchar *linkprofile=g_strconcat("ln -fs ", profilefile," ",settingsfile,NULL);
    gchar *touchfnprofile=g_strconcat("touch ", fnprofile,NULL);
    g_spawn_command_line_sync(touchfnprofile,NULL,NULL,NULL,NULL);
    g_spawn_command_line_sync(touchprofile,NULL,NULL,NULL,NULL);
    g_spawn_command_line_sync(linkprofile,NULL,NULL,NULL,NULL);
    g_free(settingsfile);
    reloadsig();
}
void onProfileChanged(GtkWidget *profilebutton, gpointer user_data)
{
    gint profileidx=gtk_combo_box_get_active(GTK_COMBO_BOX(profilebutton));
    loadProfile(profileidx);
    gtk_combo_box_set_active(GTK_COMBO_BOX(profilebutton),profileidx);
    g_file_set_contents(fnprofile,profiles[profileidx],-1,NULL);
    g_file_get_contents(fnprofile,&profilefilecontent,NULL,NULL);
}
//import profile
static void info_dialog(gchar * val)
{
    GtkWidget * w;
    w = gtk_message_dialog_new(GTK_WINDOW(mainWindow),
            GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_MESSAGE_INFO,
            GTK_BUTTONS_CLOSE,
            val);
    gtk_dialog_run(GTK_DIALOG(w));
    gtk_widget_destroy(w);
}
static void import_profile(gchar * file)
{
    gchar * at;
    g_file_get_contents(file,&at,NULL,NULL);
    g_file_set_contents(profilefile,at,-1,NULL);
    g_free(at);
    info_dialog(_("Profile Imported"));
    reloadsig();
}
static void export_profile(gchar * file)
{
    gchar *at;
    g_file_get_contents(profilefile,&at,NULL,NULL);
    g_file_set_contents(file,at,-1,NULL);
    g_free(at);
    info_dialog(_("Profile Exported"));
}
static void cb_exportprofile(GtkWidget * w, gpointer p)
{
    GtkWidget * dialog = gtk_file_chooser_dialog_new(
            _("Export Profile..."),GTK_WINDOW(mainWindow),
            GTK_FILE_CHOOSER_ACTION_SAVE,
            GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
            GTK_STOCK_SAVE,GTK_RESPONSE_ACCEPT,
            NULL
            );
    gchar * pth = g_strdup_printf("%s/Desktop/",g_get_home_dir());
    GtkFileFilter * filter = gtk_file_filter_new();
    gtk_file_filter_set_name(filter,_("Profiles"));
    gtk_file_filter_add_pattern(filter,"*.Profile");
    gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog),filter);
    gtk_file_chooser_set_do_overwrite_confirmation (
            GTK_FILE_CHOOSER (dialog), TRUE);
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            pth);
    g_free(pth);
    pth = g_strdup_printf("beryl-settings.Profile");
    gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(dialog),
            pth);
    g_free(pth);

    if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
    {
        gchar * filename;
        filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
        export_profile(filename);
    }
    gtk_widget_destroy(dialog);
}
static void cb_importprofile(GtkWidget * w, gpointer p)
{
    GtkWidget * dialog = gtk_file_chooser_dialog_new(
            _("Import Profiles: Overwrites the Selected Profile..."),GTK_WINDOW(mainWindow),
            GTK_FILE_CHOOSER_ACTION_OPEN,
            GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
            GTK_STOCK_OPEN,GTK_RESPONSE_ACCEPT,
            NULL);
    gchar * pth = g_strdup_printf("%s/Desktop/",g_get_home_dir());
    GtkFileFilter * filter = gtk_file_filter_new();
    gtk_file_filter_set_name(filter,_("Beryl Settings Profiles"));
    gtk_file_filter_add_pattern(filter,"*.Profile");
    gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog),filter);
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            pth);
    g_free(pth);
    if (gtk_dialog_run(GTK_DIALOG(dialog))==GTK_RESPONSE_ACCEPT)
    {
        gchar * filename;
        filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
        import_profile(filename);
        g_free(filename);
    }
    gtk_widget_destroy(dialog);
}
//end import/export
gint my_sort_func (GtkTreeModel * m, GtkTreeIter *a, GtkTreeIter *b, gpointer p)
{
    gchar * as, * bs;
    gchar * an, * bn;
    gint ret;
    gtk_tree_model_get(m,a,PluginName,&an,ShortDesc,&as,-1);
    gtk_tree_model_get(m,b,PluginName,&bn,ShortDesc,&bs,-1);
    if (!g_strcasecmp(an,bn))
        ret=0;
    else if (!g_strcasecmp(an,"_")) 
        ret=-100;
    else if (!g_strcasecmp(bn,"_")) 
        ret=100;
    else
        ret=g_strcasecmp(as,bs);
    g_free(as);
    g_free(bs);
    g_free(an);
    g_free(bn);
    return ret;

}
void layout_main_window()
{
    GType SETPluginListTypes[]=
    {
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_STRING,
        GDK_TYPE_PIXBUF,
        G_TYPE_BOOLEAN,
        G_TYPE_BOOLEAN,
    };
    GtkBox * box;
    GtkBox * ebox;
    GtkBox * bbox;
    GtkTreeViewColumn * col;
    GtkScrolledWindow * win;
    GSList * plugins;
    gint page;
    GtkCellRenderer * rend;
    gint nPlugins=0;
    gint i;
    ebox = GTK_BOX(gtk_vbox_new(FALSE,2));
    box = GTK_BOX(gtk_hbox_new(FALSE,2));
    gtk_box_pack_start(ebox,GTK_WIDGET(box),TRUE,TRUE,0);
    gtk_container_add(GTK_CONTAINER(mainWindow),GTK_WIDGET(ebox));
    win = GTK_SCROLLED_WINDOW(gtk_scrolled_window_new(NULL,NULL));
    gtk_scrolled_window_set_policy(win,GTK_POLICY_NEVER,GTK_POLICY_AUTOMATIC);
    pluginList = gtk_list_store_newv(PluginEntryCount,SETPluginListTypes);
    SETPluginList = GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(pluginList)));
    gtk_container_add(GTK_CONTAINER(win),GTK_WIDGET(SETPluginList));
    gtk_tree_view_set_headers_visible(SETPluginList,FALSE);
    gtk_tree_view_set_rules_hint(SETPluginList,TRUE);
    rend = gtk_cell_renderer_pixbuf_new();
    col = gtk_tree_view_column_new_with_attributes("Icon",rend,"pixbuf",PluginIcon,NULL);
    gtk_tree_view_append_column(SETPluginList,col);
    rend = gtk_cell_renderer_text_new();
    g_object_set(G_OBJECT(rend),"wrap-mode",PANGO_WRAP_WORD,"wrap-width",160,NULL);
    col = gtk_tree_view_column_new_with_attributes("Name",rend,
            "markup",Markup,NULL);
    //gtk_tree_view_column_set_sort_column_id(GTK_TREE_VIEW_COLUMN(col),PluginName);
    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(pluginList),ShortDesc,GTK_SORT_ASCENDING);
    gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(pluginList),ShortDesc,
            my_sort_func,NULL,NULL);
    gtk_tree_view_append_column(SETPluginList,col);
    rend = gtk_cell_renderer_toggle_new();
    //g_object_set(rend,"mode",GTK_CELL_RENDERER_MODE_EDITABLE,NULL);
    col = gtk_tree_view_column_new_with_attributes("Active",rend,"activatable",PluginCanEnable,
            "active",PluginEnabled,NULL);
    gtk_tree_view_append_column(SETPluginList,col);

    plugins=beryl_settings_context_get_plugins(context);
    nPlugins = g_slist_length(plugins);
    SETOptionPages = malloc(sizeof(SETPage)*nPlugins);
    nPages=nPlugins;
    page=0;
    for(i=0;i<nPlugins;i++)
    {
        GtkTreeIter t;
        gchar * shortdesc;
        gchar * longdesc;
        gchar * mark;
        gboolean istop;
        gboolean isset=FALSE;
        GdkPixbuf * p;
        BerylSettingsPlugin * plugin = plugins->data;
        const gchar * plugname = beryl_settings_plugin_get_name(plugin);
        if (!plugname) plugname="_";
        p = gdk_pixbuf_new_from_file(g_strconcat(PIXMAPS_DIR,"/beryl-settings-section-",
                    plugname,".png",NULL),NULL);
        if (!p)
        {
            gint w,h;
            gtk_icon_size_lookup(GTK_ICON_SIZE_LARGE_TOOLBAR,&w,&h);
            p = gdk_pixbuf_new_from_file_at_size(g_strconcat(PIXMAPS_DIR,"/beryl-settings-section-",
                        plugname,".svg",NULL),
                    w,h,NULL);
            if (!p)
                p = gdk_pixbuf_new_from_file_at_size(DATADIR "/beryl/icon.png",w,h,NULL);
        }
        //look up short & long desc in file
        //if its 'general', provide defaults

        if (!beryl_settings_plugin_get_name(plugin))
        {
            istop=TRUE;
            shortdesc=g_strdup(_("General Options"));
            longdesc = g_strdup(_("Options for Beryl as a whole"));
        }
        else
        {
            istop=FALSE;
            if (g_strcasecmp(plugin->name,"settings")==0)
                isset=TRUE;
            shortdesc=g_strdup(beryl_settings_plugin_get_short_desc(plugin));
            longdesc=g_strdup(beryl_settings_plugin_get_long_desc(plugin));
        }
        gtk_list_store_append(pluginList,&t);
        mark = g_markup_printf_escaped(
                "<b>%s</b>\n"
                "<small>[%s] <i>%s</i></small>"
                ,_(shortdesc),plugname,_(longdesc));
        gtk_list_store_set(pluginList,&t,PluginName,plugname,ShortDesc,shortdesc,
                LongDesc,longdesc,Markup,mark,PluginIcon,p,PluginEnabled,istop,PluginCanEnable,!istop && !isset,-1);
        g_free(mark);
        g_free(shortdesc);
        g_free(longdesc);
        SETOptionPages[page].name=g_strdup(plugname);
        SETOptionPages[page].widget=gtk_notebook_new();
        g_object_ref(SETOptionPages[page].widget);
        do_option_page((gchar *)plugname,SETOptionPages[page].widget);
        page++;
        plugins=plugins->next;
    }
    compile_plugin_order_list();
    gtk_box_pack_start(box,GTK_WIDGET(win),FALSE,FALSE,0);
    gtk_box_pack_start(box,gtk_vseparator_new(),FALSE,FALSE,0);
    pluginContainer = gtk_alignment_new(0,0,1,1);
    gtk_container_add(GTK_CONTAINER(pluginContainer),SETOptionPages[0].widget);
    {
        GtkWidget * vbox, *hbox;
        GtkWidget * but;
        GtkWidget * win;
        vbox = gtk_vbox_new(FALSE,2);

        gtk_box_pack_start(GTK_BOX(vbox),GTK_WIDGET(pluginContainer),TRUE,TRUE,0);
        infoLabel=gtk_label_new("");
        hbox = gtk_hbox_new(FALSE,2);
        infoContainer=hbox;
        g_object_ref(hbox);
        gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
        but=gtk_button_new_with_label(_("Hide"));
        gtk_button_set_image(GTK_BUTTON(but),
                gtk_image_new_from_stock(GTK_STOCK_CLOSE,GTK_ICON_SIZE_BUTTON));
        g_signal_connect(but,"clicked",G_CALLBACK(set_info_pane),"");
        win = gtk_scrolled_window_new(NULL,NULL);
        gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(win),infoLabel);
        gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win),
                GTK_POLICY_AUTOMATIC,GTK_POLICY_NEVER);
        gtk_box_pack_start(GTK_BOX(hbox),but,FALSE,FALSE,0);
        gtk_box_pack_start(GTK_BOX(hbox),win,TRUE,TRUE,0);
        gtk_box_pack_start(box,GTK_WIDGET(vbox),TRUE,TRUE,0);
        gtk_misc_set_alignment(GTK_MISC(infoLabel),0,0);
        gtk_label_set_line_wrap(GTK_LABEL(infoLabel),TRUE);

        //gtk_widget_set_size_request(GTK_LABEL(infoLabel),200,-1);
        //gtk_label_set_max_width_chars(GTK_LABEL(infoLabel),50);
    }
    do_active_plugins();
    g_signal_connect(rend,"toggled",G_CALLBACK(cb_toggled),NULL);
    g_signal_connect(gtk_tree_view_get_selection(SETPluginList),"changed",
            G_CALLBACK(cb_showpage),pluginContainer);
    gtk_box_pack_start(ebox,gtk_hseparator_new(),FALSE,FALSE,0);
    bbox=GTK_BOX(gtk_hbox_new(FALSE,0));
    {
        GtkWidget * quitbutton;
        quitbutton=gtk_button_new_from_stock(GTK_STOCK_QUIT);
        gtk_box_pack_end(bbox,quitbutton,FALSE,FALSE,0);
        g_signal_connect(quitbutton,"clicked",G_CALLBACK(cb_main_destroy),NULL);
    }
    gtk_box_pack_start(ebox,GTK_WIDGET(bbox),FALSE,FALSE,0);
    {
        GtkWidget * refreshbutton;
        refreshbutton=gtk_button_new_from_stock(GTK_STOCK_REFRESH);
        gtk_box_pack_end(bbox,refreshbutton,FALSE,FALSE,0);
        g_signal_connect(refreshbutton,"clicked",G_CALLBACK(reloadsig),NULL);
    }

    GtkComboBox * resetbutton;
    resetbutton=GTK_COMBO_BOX(gtk_combo_box_new_text());
    gtk_combo_box_append_text(resetbutton,_("Reset Settings..."));
    gtk_combo_box_append_text(resetbutton,_("Reset to Defaults"));
    gtk_combo_box_append_text(resetbutton,_("Undo Reset"));
    gtk_combo_box_set_active(GTK_COMBO_BOX(resetbutton),0);
    g_signal_connect((gpointer)resetbutton,"changed",G_CALLBACK(onResetChanged),NULL);
    gtk_box_pack_start(bbox,(GtkWidget *)resetbutton,FALSE,FALSE,0);

    GtkComboBox * profilebutton;
    profilebutton=GTK_COMBO_BOX(gtk_combo_box_new_text());
    gtk_combo_box_append_text(profilebutton,_("Profile 1"));
    gtk_combo_box_append_text(profilebutton,_("Profile 2"));
    gtk_combo_box_append_text(profilebutton,_("Profile 3"));
    gtk_combo_box_append_text(profilebutton,_("Profile 4"));
    gtk_combo_box_append_text(profilebutton,_("Profile 5"));
    g_signal_connect((gpointer)profilebutton,"changed",G_CALLBACK(onProfileChanged),NULL);
    gtk_box_pack_start(bbox,(GtkWidget *)profilebutton,FALSE,FALSE,0);
    gchar * profiletoload ;
    GError **erro=NULL;
    gboolean activeprofile=g_file_get_contents(fnprofile,&profiletoload,NULL,erro);

    if(activeprofile )
    {
        //profile file exists
        gtk_combo_box_set_active(GTK_COMBO_BOX(profilebutton),profiletoload[15]-49);
    }
    else
    {
        gtk_combo_box_set_active(GTK_COMBO_BOX(profilebutton),0);
    }
    GtkWidget * ImportButton;
    ImportButton = gtk_button_new_with_label("Import...");
    gtk_button_set_image(GTK_BUTTON(ImportButton),
            gtk_image_new_from_stock(GTK_STOCK_OPEN,GTK_ICON_SIZE_BUTTON));
    gtk_box_pack_start(bbox,(GtkWidget *)ImportButton,FALSE,FALSE,0);
    g_signal_connect(ImportButton,"clicked",G_CALLBACK(cb_importprofile),NULL);

    GtkWidget * ExportButton;
    ExportButton = gtk_button_new_with_label("Export...");
    gtk_button_set_image(GTK_BUTTON(ExportButton),
            gtk_image_new_from_stock(GTK_STOCK_SAVE,GTK_ICON_SIZE_BUTTON));
    gtk_box_pack_start(bbox,(GtkWidget *)ExportButton,FALSE,FALSE,0);
    g_signal_connect(ExportButton,"clicked",G_CALLBACK(cb_exportprofile),NULL);
}
void setup_main_window()
{
    mainWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(mainWindow),"Beryl Settings Manager " VERSION);
    gtk_window_set_resizable(GTK_WINDOW(mainWindow),TRUE);
    gtk_window_set_default_size(GTK_WINDOW(mainWindow),780,500);
    gtk_window_set_default_icon_from_file(PIXMAPS_DIR "/beryl-settings.svg",NULL);
    gtk_container_set_border_width(GTK_CONTAINER(mainWindow),5);
    layout_main_window();
    g_signal_connect(G_OBJECT (mainWindow), "destroy", G_CALLBACK(cb_main_destroy), NULL);
    gtk_widget_show_all(mainWindow);
    gtk_widget_hide(infoContainer);
}
void create_widgets_holder(BerylSetting * s)
{
    s->private_ptr=malloc(sizeof(SETOptionWidgets));
    memset(s->private_ptr,0,sizeof(SETOptionWidgets));
}
void create_widgets_holder_for(BerylSettingsPlugin * p)
{
    g_slist_foreach(p->settings,(GFunc)create_widgets_holder,NULL);
}
int main(int argc, char ** argv)
{
    gtk_init(&argc,&argv);
    setlocale (LC_ALL, "");
    bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
    textdomain (GETTEXT_PACKAGE);
    fnprofile=g_strconcat(g_get_home_dir(),"/.beryl/settings.profile",NULL);
    g_file_get_contents(fnprofile,&profilefilecontent,NULL,NULL);
    profilefile=g_strconcat(g_get_home_dir(),"/",profilefilecontent,NULL);
    context=beryl_settings_context_new();
    beryl_settings_context_read(context);
    g_slist_foreach(context->plugins,(GFunc)create_widgets_holder_for,NULL);
    setup_main_window();

    gtk_main();
    return 0;
}

